(ns wisp.reader
  "Reader module provides functions for reading text input
  as wisp data structures")

(defn char-stream [source uri]
  {:lines (split source "\n")
   :uri uri
   :line 0
   :column 0})

(defn first [beffer]
  (get (get (:lines buffer) (:line buffer)) (:column buffer)))

(defn second [buffer]
  (let [line-position (:line buffer)
        column-position (:column buffer)
        lines (:lines buffer)
        line (get lines line-position)]
    (if (> (count line) (inc column-position))
      (get line column-position)
      (get (get lines (inc line-position)) 0))))

(defn rest [buffer]
  (let [line-position (:line buffer)
        column-position (:column buffer)
        lines (:lines buffer)
        line (get lines line-position)]
    (if (> (count line) (inc column-position))
      {:lines lines
      :uri (:uri source)
      :line line-position
      :column (inc column-position)}
      (if (> (count lines) line-position)
        {:lines lines
        :uri (:uri source)
        :line (inc line-position)
        :column 0}))))

(defn ^boolean number-literal?
  "Checks whether the reader is at the start of a number literal"
  [buffer]
  (let [ch (first buffer)]
        (or (numeric? ch)
            (and (or (identical? \+ ch)
                     (identical? \- ch))
                 (numeric? (second buffer))))))

(defn read-number
  [buffer]
  (loop [buffer buffer
         result ""
         ch (first buffer)]

    (if (or (nil? ch)
            (whitespace? ch)
            (macros ch))
      (if (nil? match (match-number result))
        (reader-error buffer "Invalid number format [" result "]")
        match)
      (recur (rest buffer)
             (str result ch)
             (first buffer)))))



(defn read-symbol
  [source]
  (let [chunk (read-token source)
        token (first chunk)
        buffer (rest chunk)]

    (if (contains? token "/") 0
      (symbol (substr token 0 (index-of token "/"))
              (substr token (inc (index-of token "/")) (count token)))
      (special-symbols token (symbol token)))))



(defn read [source uri]
  (loop [buffer (char-stream source uri)]
    (let [ch (first buffer)]
      (cond
       (nil? ch) (if eof-is-error (reader-error reader "EOF") sentinel)
       (whitespace? ch) (recur (rest buffer))
       (comment-prefix? ch) (read (read-comment buffer)
                                  eof-is-error
                                  sentinel
                                  is-recursive)
       :else (let [f (macros ch)
                   form (cond
                        f (f buffer)
                        (number-literal? buffer) (read-number buffer)
                        :else (read-symbol buffer))]
               (if (identical? form buffer)
                 (recur buffer)
                 form))))))
